import textwrap
from copy import deepcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Tuple, Union

import networkx as nx

from cmstp.core.logger import Logger
from cmstp.utils.command import Command, CommandKind
from cmstp.utils.common import PACKAGE_CONFIG_PATH, resolve_package_path
from cmstp.utils.patterns import PatternCollection
from cmstp.utils.tasks import (
    TASK_ARGS_CUSTOM,
    TASK_ARGS_DEFAULT,
    TASK_PROPERTIES_CUSTOM,
    TASK_PROPERTIES_DEFAULT,
    ResolvedTask,
    TaskDictCollection,
    get_invalid_tasks_from_task_dict_collection,
    print_expected_task_fields,
)
from cmstp.utils.yaml import load_yaml, overlay_dicts


# TODO: Extract (into funcs) and order tests, e.g. check structure, check script/function existence, check args, check dependencies, check supercedes, etc.
@dataclass
class TaskProcessor:
    """Processes task configurations and resolves tasks to be run."""

    # fmt: off
    logger:           Logger             = field(repr=False)
    config_file:      Optional[Path]     = field()
    config_directory: Path               = field()
    custom_tasks:     List[str]          = field()

    enable_all:       bool               = field(init=False, repr=False, default=False)
    enable_deps:      bool               = field(init=False, repr=False, default=False)
    resolved_tasks:   List[ResolvedTask] = field(init=False, repr=False, default=None)
    default_config:   Path               = field(
        init=False,
        repr=False,
        default=PACKAGE_CONFIG_PATH / "default.yaml",
    )
    # fmt: on

    def __post_init__(self):
        # (Internal) Check default config file
        default_config = self.check_default_config()

        # Check custom config file
        if self.config_file is not None:
            self.config_file = resolve_package_path(self.config_file)
            if self.config_file is None:
                self.logger.fatal(
                    f"Custom config file not found: {self.config_file}"
                )
            custom_config = self.check_custom_config()
        else:
            self.logger.debug(
                "Tasks have been specified directly and no config "
                "file is specified, so only those tasks will be run"
            )
            custom_config = {}
            for task_name in self.custom_tasks:
                custom_config[task_name] = {"enabled": True}

        # Merge default and custom config
        tasks = overlay_dicts(
            default_config, custom_config, allow_default=True
        )

        # Disable tasks in custom config that are not in default config
        for task_name in custom_config:
            if task_name not in default_config:
                self.logger.warning(
                    f"Task '{task_name}' in custom config is disabled "
                    f"because it is not defined in the default config"
                )
                tasks[task_name]["enabled"] = False

        # Fill in missing optional fields with defaults (including for default config)
        tasks = self.fill_missing_fields(tasks, self.enable_all)

        # Resolve dependencies (disable tasks that depend on disabled ones)
        tasks = self.check_dependencies(tasks)

        # Process args (give defaults and disable tasks with invalid args)
        tasks = self.check_args(tasks)

        # Handle conflicting/superceding tasks
        tasks = self.check_conflicts(tasks)

        # Prepend config directory to config file paths
        tasks = self.prepend_config_directory(tasks)

        # Count enabled tasks
        if self.count_tasks(tasks) == 0:
            self.logger.fatal(
                "There are no enabled tasks (anymore). Nothing to do"
            )
        else:
            self.logger.debug(
                f"Final enabled tasks: {[name for name, task in tasks.items() if task['enabled']]}"
            )

        # Create logging directory
        self.logger.create_log_dir()

        # Convert to ResolvedTask list
        self.resolved_tasks = []
        for task_name, task in tasks.items():
            if not task["enabled"]:
                continue

            resolved_task = ResolvedTask(
                name=task_name,
                command=Command(task["script"], task["function"]),
                config_file=task["config_file"],
                depends_on=tuple(task["depends_on"]),
                args=tuple(task["_resolved_args"]),
            )
            self.resolved_tasks.append(resolved_task)

    def prepend_config_directory(
        self, tasks: TaskDictCollection
    ) -> TaskDictCollection:
        # TODO: Move to better location, e.g. after overlay?
        """
        Prepend config directory to config file paths in tasks.
        """
        for task_name, task in tasks.items():
            if task["config_file"] is not None:
                config_file = (
                    self.config_directory / task["config_file"]
                ).resolve()

                if not config_file.exists():
                    self.logger.fatal(
                        f"Config file '{config_file}' for "
                        f"task '{task_name}' does not exist"
                    )
                else:
                    task["config_file"] = str(config_file)

        return tasks

    def fill_missing_fields(
        self,
        tasks: TaskDictCollection,
        enable_all: bool = False,
        include_default: bool = True,
        include_custom: bool = True,
    ) -> TaskDictCollection:
        """
        Fill in missing optional fields with defaults.
        """
        expected_properties = dict()
        expected_args = dict()
        if include_default:
            expected_properties |= TASK_PROPERTIES_DEFAULT
            expected_args |= TASK_ARGS_DEFAULT
        if include_custom:
            expected_properties |= TASK_PROPERTIES_CUSTOM
            expected_args |= TASK_ARGS_CUSTOM

        filled_tasks = deepcopy(tasks)
        for task_name, task in filled_tasks.items():
            # Task properties
            for task_field, default in expected_properties.items():
                # NOTE: bool() = False
                called_default = (
                    default[0]() if callable(default[0]) else default[0]
                )
                if task_field not in task:
                    if task_field == "enabled":
                        task[task_field] = enable_all
                        if task_name.startswith("uninstall"):
                            # Uninstall tasks are disabled by default for now - TODO: Change as soon as separate "uninstall" entrypoint is added
                            task[task_field] = False
                    elif task_field == "config_file":
                        task[task_field] = "default"
                    else:
                        task[task_field] = called_default
                else:
                    if not (
                        (None in default and task[task_field] is None)
                        or any(
                            isinstance(task[task_field], t)
                            for t in default
                            if t is not None
                        )
                    ):
                        self.logger.warning(
                            f"Disabling task '{task_name}' because task property '{task_field}' has incorrect "
                            f"type (expected: {default}, got: {type(task[task_field]).__name__})"
                        )
                        task[task_field] = called_default

            # Task args
            if "args" not in task:
                task["args"] = {}
            elif not isinstance(task["args"], dict):
                self.logger.warning(
                    f"Disabling task '{task_name}' because 'args' field is not a dict"
                )
                task["enabled"] = False
                task["args"] = {}
            for arg_field, default in expected_args.items():
                # NOTE: bool() = False
                called_default = (
                    default[0]() if callable(default[0]) else default[0]
                )
                if arg_field not in task["args"]:
                    task["args"][arg_field] = called_default
                else:
                    if not (
                        (None in default and task["args"][arg_field] is None)
                        or any(
                            isinstance(task["args"][arg_field], t)
                            for t in default
                            if t is not None
                        )
                    ):
                        self.logger.warning(
                            f"Disabling task '{task_name}' because task arg '{arg_field}' has incorrect "
                            f"type (expected: {default}, got: {type(task['args'][arg_field]).__name__})"
                        )
                        task["args"][arg_field] = called_default

        return filled_tasks

    def check_default_config(self) -> TaskDictCollection:
        """
        Check that the default config file is valid.
        """

        def fatal(msg: str, task_name: Optional[str] = None):
            """Helper to log fatal errors"""
            self.logger.fatal(
                f"Error in default config file {self.default_config}: {f'{task_name}:' if task_name else ''} {msg}"
            )

        # Check file exists and is not empty
        default_config = load_yaml(self.default_config)
        if default_config is None:
            fatal("File does not exist or is not valid YAML")
        if not default_config:
            fatal("File is empty")
        if not isinstance(default_config, dict):
            fatal(
                "File does not define a dict, but a "
                + type(default_config).__name__
            )

        # Remove helpers (start with '_') that may otherwise be picked up as tasks
        default_config = {
            k: v for k, v in default_config.items() if not k.startswith("_")
        }

        # Check structure (incl. types)
        invalid_tasks = get_invalid_tasks_from_task_dict_collection(
            default_config
        )
        if invalid_tasks:
            fatal_msg = textwrap.dedent(
                f"""\
                Some tasks have extra or missing fields, or use incorrect types: {invalid_tasks}
                [cyan]Required structure for tasks in the default config[/cyan]:
            """
            )
            fatal(
                fatal_msg
                + print_expected_task_fields(
                    TASK_ARGS_DEFAULT, TASK_PROPERTIES_DEFAULT
                )
            )

        # Check everything else
        existing_scripts = set()
        for task_name, task in default_config.items():
            # Check 'script' field
            if task["script"] is None:
                fatal(
                    "Script is either null or uses a package that can't be found",
                    task_name,
                )
            if not Path(task["script"]).exists():
                fatal(f"'{task['script']}' script does not exist", task_name)

            # Check for duplicate (script, function) pairs
            script_command = Command(task["script"], task["function"])
            if script_command in existing_scripts:
                fatal(
                    f"Duplicate (script, function) pair: {script_command}",
                    task_name,
                )
            else:
                existing_scripts.add(script_command)

            # Check 'function' field
            if task["function"] is not None:
                # Get script info
                script_kind = script_command.kind
                with open(task["script"]) as f:
                    lines = f.readlines()

                # Find function in script
                function_pattern = PatternCollection[
                    script_kind.name
                ].patterns["blocks"]["FUNCTION"]
                function_matches = [
                    match.groups()
                    for line in lines
                    if (match := function_pattern.search(line.strip()))
                ]
                if script_kind == CommandKind.PYTHON:  # Also capture args
                    function_names = [name for name, _ in function_matches]
                else:  # No args are captured in bash
                    function_names = [name for name, in function_matches]
                if task["function"] not in function_names:
                    fatal(
                        f"'{task['function']}' function not found in script "
                        f"{task['script']}\nAvailable functions: {function_names}",
                        task_name,
                    )

                # If Python, check function definition only captures '*args'
                elif script_kind == CommandKind.PYTHON:
                    # Test if the function has only *args
                    function_args = [args for _, args in function_matches]
                    args = function_args[
                        function_names.index(task["function"])
                    ]
                    arg_list = [
                        a.strip() for a in args.split(",") if a.strip()
                    ]
                    if not (len(arg_list) == 1 and arg_list[0] == "*args"):
                        fatal(
                            f"'{task['function']}' function in script "
                            f"{task['script']} must ONLY capture '*args' as "
                            f"an argument, if it is to be used as a task",
                            task_name,
                        )

            # Check 'depends_on' field (must refer to existing tasks)
            for dep in task["depends_on"]:
                if dep not in default_config:
                    fatal(f"'{dep}' dependency task does not exist", task_name)

            # Check default args are allowed
            wrong_args, is_allowed = self.check_allowed(
                task["args"]["allowed"], task["args"]["default"]
            )
            if not is_allowed:
                fatal(
                    f"Some default args ({wrong_args}) are not in allowed args {task['args']['allowed']}",
                    task_name,
                )

        self.logger.debug("Default config file is valid")
        return default_config

    def check_custom_config(self) -> TaskDictCollection:
        """
        Check that the custom config file is valid.
        """

        def warning(msg: str):
            """Helper to log warning messages"""
            self.logger.warning(
                f"Warning in custom config file {self.config_file}: {msg}"
            )

        def fatal(msg: str):
            """Helper to log fatal errors"""
            self.logger.fatal(
                f"Error in custom config file {self.config_file}: {msg}"
            )

        # Check file exists and is not empty
        custom_config = load_yaml(self.config_file)
        if custom_config is None:
            fatal("File does not exist or is not valid YAML")
        if not custom_config:
            fatal("File is empty")
        if not isinstance(custom_config, dict):
            fatal(
                "File does not define a dict, but a "
                + type(custom_config).__name__
            )

        # Check for "enable_all" parameter
        enable_all = False
        enable_all_key = "enable_all"
        if enable_all_key in custom_config:
            enable_all = custom_config.pop(enable_all_key)
            if not isinstance(enable_all, bool):
                warning(
                    f"Ignoring '{enable_all_key}' value - must be "
                    f"a boolean, not a {type(enable_all).__name__}"
                )
                enable_all = False
        self.enable_all = enable_all

        # Check for "enable_deps" parameter
        enable_deps = False
        enable_deps_key = "enable_deps"
        if enable_deps_key in custom_config:
            enable_deps = custom_config.pop(enable_deps_key)
            if not isinstance(enable_deps, bool):
                warning(
                    f"Ignoring '{enable_deps_key}' value - must be "
                    f"a boolean, not a {type(enable_deps).__name__}"
                )
                enable_deps = False
        self.enable_deps = enable_deps

        # Add custom tasks if specified directly
        if self.custom_tasks is not None:
            for task_name in self.custom_tasks:
                # NOTE: If both custom config file and direct tasks are specified,
                #       the directly specified tasks take precedence in enabling
                if task_name not in custom_config:
                    custom_config[task_name] = {"enabled": True}
                else:
                    custom_config[task_name]["enabled"] = True

        # Add defaults for missing optional fields. Used to check structure of custom config tasks
        checked_custom_config = self.fill_missing_fields(
            custom_config, enable_all=enable_all, include_default=False
        )

        # Check structure (incl. types)
        invalid_tasks = get_invalid_tasks_from_task_dict_collection(
            checked_custom_config, include_default=False, include_custom=True
        )
        if invalid_tasks:
            warning_msg = textwrap.dedent(
                f"""\
                Some tasks that have extra fields or are trying to override default fields are disabled: {invalid_tasks}
                [cyan]Required structure for tasks in the custom config[/cyan] (all fields optional):
            """
            )
            warning(
                warning_msg
                + print_expected_task_fields(
                    TASK_ARGS_CUSTOM, TASK_PROPERTIES_CUSTOM
                )
            )
            for task in invalid_tasks:
                checked_custom_config[task]["enabled"] = False

        # If enable_all is off, check if any tasks are still enabled
        if not enable_all and self.count_tasks(checked_custom_config) == 0:
            fatal("No valid tasks defined")

        self.logger.debug("Custom config file is valid")
        return checked_custom_config

    @staticmethod
    def check_allowed(
        allowed_args: Optional[List[str]], args: Union[str, List[str]]
    ) -> Tuple[List[str], bool]:
        """
        Check if an argument is in the allowed list.
        If allowed_args is None, any argument is allowed.
        """

        def _check_single_allowed(allowed_args: List[str], arg: str) -> bool:
            """Check if a single argument is allowed, supporting wildcard '*'"""
            for allowed in allowed_args:
                if allowed.endswith("*"):
                    if arg.startswith(allowed[:-1]):
                        return True
                elif arg == allowed:
                    return True
            return False

        if allowed_args is None:
            return [], True
        if isinstance(args, str):
            is_allowed = _check_single_allowed(allowed_args, args)
            wrong_args = [] if is_allowed else [args]
            return wrong_args, is_allowed
        else:  # List[str]
            wrong_args = [
                arg
                for arg in args
                if not _check_single_allowed(allowed_args, arg)
            ]
            return wrong_args, not wrong_args

    def check_args(self, tasks: TaskDictCollection) -> TaskDictCollection:
        """
        Check task arguments for validity (if they are in the allowed list).
        """
        for task_name, task in tasks.items():
            if task["enabled"] is False:
                continue

            # Desired args (duplicates are allowed)
            task["_resolved_args"] = task["args"]["custom"]
            if not task["args"]["override_default"]:
                task["_resolved_args"] += task["args"]["default"]

            # Allowed args - if 'None', then any arg is allowed
            wrong_args, is_allowed = self.check_allowed(
                task["args"]["allowed"], task["_resolved_args"]
            )
            if not is_allowed:
                # Disable task
                task["enabled"] = False
                self.logger.warning(
                    f"Task '{task_name}' was disabled because it uses "
                    f"unallowed args {wrong_args} (allowed: {task['args']['allowed']})"
                )

        self.logger.debug("Checked task arguments for validity")
        return tasks

    def check_dependencies(
        self, tasks: TaskDictCollection
    ) -> TaskDictCollection:
        """
        Mutates `tasks` in place:
        - Ensures all dependencies exist
        - Raises fatal error if dependency graph has cycles
        - If `self.enable_deps` is True:
            Enables all dependencies of enabled tasks (recursively)
        - Otherwise:
            Disables any task whose dependencies are disabled
        """
        # Build dependency graph
        G = nx.DiGraph()
        for task_name, task in tasks.items():
            G.add_node(task_name)
            for dep in task["depends_on"]:
                if dep not in tasks:
                    self.logger.fatal(
                        f"Dependency '{dep}' for '{task_name}' "
                        f"not found - needs to be defined in config"
                    )
                G.add_edge(dep, task_name)

        # Check for cycles
        if not nx.is_directed_acyclic_graph(G):
            cycles = list(nx.simple_cycles(G))
            self.logger.fatal(f"Dependency graph has cycles between: {cycles}")

        # Topological resolution
        for node in nx.topological_sort(G):
            task = tasks[node]

            if self.enable_deps:
                # Recursively enable dependencies
                if task["enabled"]:
                    # Recursively enable all dependencies of this enabled task
                    for dep in nx.ancestors(G, node):
                        if not tasks[dep]["enabled"]:
                            tasks[dep]["enabled"] = True
                            self.logger.info(
                                f"Enabling dependency '{dep}' "
                                f"because '{node}' is enabled and enable_deps=True"
                            )

            else:
                if not task["enabled"]:
                    continue
                # Disable task if any dependencies are disabled
                deps = list(G.predecessors(node))
                disabled_deps = [d for d in deps if not tasks[d]["enabled"]]
                if disabled_deps:
                    task["enabled"] = False
                    self.logger.warning(
                        f"Task '{node}' was disabled because it depends on "
                        f"disabled tasks: {disabled_deps}"
                    )

        self.logger.debug("Checked task dependencies")
        return tasks

    def check_conflicts(self, tasks: TaskDictCollection) -> TaskDictCollection:
        """
        Disable tasks that are conflicting with superceding tasks.
        """
        # First, collect all tasks that are enabled and have 'supercedes' field
        superceding_tasks = {
            task_name: task
            for task_name, task in tasks.items()
            if task["enabled"] and task["supercedes"] is not None
        }

        # Disable conflicting tasks
        for task_name, task in superceding_tasks.items():
            for conflicted_task in task["supercedes"]:
                if tasks[conflicted_task]["enabled"]:
                    tasks[conflicted_task]["enabled"] = False
                    self.logger.debug(
                        f"Disabling task '{conflicted_task}' because it "
                        f"conflicts with superceding task '{task_name}'"
                    )

        self.logger.debug("Checked for conflicting/superceding tasks")
        return tasks

    @staticmethod
    def count_tasks(tasks: TaskDictCollection) -> int:
        """Count the number of enabled tasks."""
        count = 0
        for task in tasks.values():
            if task.get("enabled", False):
                count += 1
        return count



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, TypedDict

from cmstp.utils.command import Command

FieldTypeDict = Dict[str, List[Optional[type]]]

# Required in default config - "'script': None" is actually not allowed, but defaults to None if nonexistent package is used
TASK_ARGS_DEFAULT: FieldTypeDict = {
    "allowed": [None, list],
    "default": [list],
}
TASK_PROPERTIES_DEFAULT: FieldTypeDict = {
    "description": [str],  # TODO: Use and add pytest
    "script": [None, str],
    "function": [None, str],
    "config_file": [None, str],  # TODO: Use and add pytest
    "depends_on": [list],
    "supercedes": [None, list],  # TODO: Use and add pytest
}

# Optional in custom config
TASK_ARGS_CUSTOM: FieldTypeDict = {
    "custom": [list],
    "override_default": [bool],
}
TASK_PROPERTIES_CUSTOM: FieldTypeDict = {
    "enabled": [bool],
    "config_file": [None, str],  # TODO: Use and add pytest
}

HARDWARE_SPECIFIC_TASKS = ["install-nvidia-driver", "install-cuda"]


def print_expected_task_fields(
    args_types: FieldTypeDict, keys_types: FieldTypeDict
) -> str:
    """
    Returns a YAML-like string with a top-level task name, first-level
    properties and and second-level args, along with their expected types.
    """

    def format_value(value):
        formatted_items = []
        for v in value:
            if isinstance(v, type):
                formatted = v.__name__
            elif v is None:
                formatted = "null"
            else:
                formatted = str(v)
            formatted_items.append(formatted)

        return ", ".join(formatted_items)

    # Prepare key width for alignment (including indented args keys)
    all_keys = list(keys_types.keys()) + [f"  {k}" for k in args_types.keys()]
    max_key_len = max(len(k) for k in all_keys) + 2

    lines = ["<task-name>:"]
    indent = "  "  # base indent for content under task

    # Add top-level keys
    for key, value in keys_types.items():
        lines.append(
            f"{indent}{key}:{' ' * (max_key_len - len(key))}{format_value(value)}"
        )

    # Add args section
    if args_types:
        lines.append(f"{indent}args:")
        for key, value in args_types.items():
            indented_key = f"{indent}  {key}"
            lines.append(
                f"{indented_key}:{' ' * (max_key_len - len(f'  {key}'))}{format_value(value)}"
            )

    return "\n".join(lines)


def check_structure(obj: Any, expected: FieldTypeDict) -> bool:
    """Check if an object matches its expected structure"""
    if not isinstance(obj, dict):
        return False
    if set(obj.keys()) != set(expected.keys()):
        return False
    for key, types in expected.items():
        if None in types and obj[key] is None:
            continue
        if type(obj[key]) not in types:
            return False
    return True


class ArgsDict(TypedDict):
    # fmt: off
    allowed:          Optional[List[str]]
    custom:           List[str]
    default:          List[str]
    override_default: bool
    # fmt: on


def is_args_dict(
    obj: Any, include_default: bool = True, include_custom: bool = False
) -> bool:
    """Check if an object is a valid ArgsDict"""
    expected_args = dict()
    if include_default:
        expected_args |= TASK_ARGS_DEFAULT
    if include_custom:
        expected_args |= TASK_ARGS_CUSTOM

    return check_structure(obj, expected_args)


class TaskDict(TypedDict):
    # fmt: off
    enabled:        bool
    description:    str
    script:         str
    function:       Optional[str]
    config_file:    Optional[str]
    depends_on:     List[str]
    supercedes:     Optional[List[str]]
    args:           ArgsDict

    _resolved_args: List[str]
    # fmt: on


def is_task_dict(
    obj: Any, include_default: bool = True, include_custom: bool = False
) -> bool:
    """Check if an object is a valid TaskDict"""
    expected_keys = dict()
    if include_default:
        expected_keys |= TASK_PROPERTIES_DEFAULT
    if include_custom:
        expected_keys |= TASK_PROPERTIES_CUSTOM

    if not isinstance(obj, dict):
        return False
    obj_noargs = {k: v for k, v in obj.items() if k != "args"}
    obj_args = obj.get("args")

    return check_structure(obj_noargs, expected_keys) and is_args_dict(
        obj_args,
        include_default=include_default,
        include_custom=include_custom,
    )


TaskDictCollection = Dict[str, TaskDict]


def get_invalid_tasks_from_task_dict_collection(
    obj: Dict[Any, Any],
    include_default: bool = True,
    include_custom: bool = False,
) -> Optional[List[str]]:
    """
    Check if an object is a valid collection of TaskDicts.
    Returns a list of invalid task names.
    If the entire object is not the required input type (dict), returns None.

    NOTE: This allows empty dicts (no tasks) as valid input
    """
    if not isinstance(obj, dict):
        return None

    invalid_tasks = []
    for key, value in obj.items():
        if not isinstance(key, str) or not is_task_dict(
            value,
            include_default=include_default,
            include_custom=include_custom,
        ):
            invalid_tasks.append(key)
    return invalid_tasks


@dataclass(frozen=True)
class ResolvedTask:
    """Represents a resolved task with its name, command, dependencies, and arguments."""

    # fmt: off
    name:        str           = field()
    command:     Command       = field()
    config_file: Optional[str] = field(default=None)
    depends_on:  Tuple[str]    = field(default_factory=tuple)
    args:        Tuple[str]    = field(default_factory=tuple)
    # fmt: on



import re
from pathlib import Path
from unittest.mock import patch

import pytest
from _pytest.capture import CaptureFixture

from cmstp.core.logger import Logger
from cmstp.core.task_processor import TaskProcessor
from cmstp.utils.common import (
    PACKAGE_CONFIG_PATH,
    PACKAGE_TESTS_PATH,
    generate_random_path,
)


def get_empty_yaml_file() -> str:
    """Create and return a path to an empty temporary YAML file"""
    return str(generate_random_path(suffix=".yaml", create=True))


def get_missing_yaml_file() -> str:
    """Generate and return a path to a non-existent YAML file"""
    return str(generate_random_path(suffix=".yaml", create=False))


#################################################################################################################
################################################ Package Configs ################################################
#################################################################################################################


def test_package_configs():
    """See if the package default and custom configs are valid"""
    dummy_logger = Logger(parser=None)
    dummy_path = Path()
    TaskProcessor(
        logger=dummy_logger,
        config_file=PACKAGE_CONFIG_PATH / "enabled.yaml",
        config_dir=dummy_path,
    )


#################################################################################################################
############################################# Default Config Checks #############################################
#################################################################################################################


class TestDefaultConfig:
    """Group default config tests in a class for better organization"""

    def setup_method(self):
        """Setup that runs before each test method"""
        self.config_dir: Path = (
            PACKAGE_TESTS_PATH / "config" / "test_task_processor" / "default"
        )
        self.fields_message = (
            "Some tasks have extra or missing fields, or use incorrect types"
        )

    def _check_config(
        self, capsys: CaptureFixture, config_file: str, expected_err: str
    ) -> None:
        with patch.object(
            TaskProcessor, "default_config", self.config_dir / config_file
        ):
            with pytest.raises(SystemExit) as exc_info:
                TaskProcessor(
                    logger=Logger(parser=None),  # Dummy logger
                    config_file=self.config_dir / ".." / "valid_custom.yaml",
                    config_dir=Path(),  # Dummy path
                )

        captured_err = capsys.readouterr().err
        captured_err_clean = re.sub(r"\s+", " ", captured_err).strip()

        assert exc_info.value.code == 1
        assert expected_err in captured_err_clean

    def test_missing_config(self, capsys: CaptureFixture):
        """See if a missing default config is handled correctly"""
        self._check_config(
            capsys, get_missing_yaml_file(), "File does not exist"
        )

    def test_empty_config(self, capsys: CaptureFixture):
        """See if an empty default config is handled correctly"""
        self._check_config(capsys, get_empty_yaml_file(), "File is empty")

    def test_duplicate_task(self, capsys: CaptureFixture):
        """See if a default config with duplicate tasks is handled correctly"""
        self._check_config(
            capsys,
            "test_duplicate_task.yaml",
            "Duplicate (script, function) pair",
        )

    ### Properties ####

    def test_missing_task_property(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_missing_task_property.yaml", self.fields_message
        )

    def test_extra_task_property(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_extra_task_property.yaml", self.fields_message
        )

    def test_wrong_task_property_type(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_wrong_task_property_type.yaml", self.fields_message
        )

    def test_missing_script(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_missing_script.yaml", "script does not exist"
        )

    def test_script_with_nonexistent_package(self, capsys: CaptureFixture):
        self._check_config(
            capsys,
            "test_script_with_nonexistent_package.yaml",
            "Script is either null or uses a package that can't be found",
        )

    def test_nonexistent_function(self, capsys: CaptureFixture):
        self._check_config(
            capsys,
            "test_nonexistent_function.yaml",
            "function not found in script",
        )

    def test_wrong_function_definition(self, capsys: CaptureFixture):
        self._check_config(
            capsys,
            "test_wrong_function_definition.yaml",
            "must ONLY capture '*args' as an argument",
        )

    def test_missing_dependency(self, capsys: CaptureFixture):
        self._check_config(
            capsys,
            "test_missing_dependency.yaml",
            "dependency task does not exist",
        )

    def test_cyclic_dependency(self, capsys: CaptureFixture):
        self._check_config(
            capsys,
            "test_cyclic_dependency.yaml",
            "Dependency graph has cycles between",
        )

    def test_wrong_args_type(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_wrong_args_type.yaml", self.fields_message
        )

    ### Args ####

    def test_missing_task_arg(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_missing_task_arg.yaml", self.fields_message
        )

    def test_extra_task_arg(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_extra_task_arg.yaml", self.fields_message
        )

    def test_wrong_task_arg_type(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_wrong_task_arg_type.yaml", self.fields_message
        )

    def test_unallowed_arg(self, capsys: CaptureFixture):
        self._check_config(
            capsys, "test_unallowed_arg.yaml", "not in allowed args"
        )


#################################################################################################################
############################################# Custom Config Checks ##############################################
#################################################################################################################


class TestCustomConfig:
    """Group custom config tests in a class for better organization"""

    def setup_method(self):
        """Setup that runs before each test method"""
        self.config_dir: Path = (
            PACKAGE_TESTS_PATH / "config" / "test_task_processor" / "custom"
        )
        self.fields_message = "Some tasks that have extra fields or are trying to override default fields are disabled"

    def _check_config_error(
        self, capsys: CaptureFixture, config_file: str, expected_err: str
    ) -> None:
        with patch.object(
            TaskProcessor,
            "default_config",
            self.config_dir / ".." / "valid_default.yaml",
        ):
            with pytest.raises(SystemExit) as exc_info:
                # Use dummy config paths, as they are not used in the TaskProcessor when checking the default config
                TaskProcessor(
                    logger=Logger(parser=None),  # Dummy logger
                    config_file=self.config_dir / config_file,
                    config_dir=Path(),  # Dummy path
                )

        captured_err = capsys.readouterr().err
        captured_err_clean = re.sub(r"\s+", " ", captured_err).strip()

        assert exc_info.value.code == 1
        assert expected_err in captured_err_clean

    def _check_config_warning(
        self, capsys: CaptureFixture, config_file: str, expected_warn: str
    ) -> None:
        with patch.object(
            TaskProcessor,
            "default_config",
            self.config_dir / ".." / "valid_default.yaml",
        ):
            task_processor = TaskProcessor(
                logger=Logger(parser=None),  # Dummy logger
                config_file=self.config_dir / config_file,
                config_dir=Path(),  # Dummy path
            )

        captured_warn = capsys.readouterr().out
        captured_warn_clean = re.sub(r"\s+", " ", captured_warn).strip()

        assert expected_warn in captured_warn_clean

    def test_missing_config(self, capsys: CaptureFixture):
        """See if a missing default config is handled correctly"""
        self._check_config_error(
            capsys, get_missing_yaml_file(), "File does not exist"
        )

    def test_empty_config(self, capsys: CaptureFixture):
        """See if an empty default config is handled correctly"""
        self._check_config_error(
            capsys, get_empty_yaml_file(), "File is empty"
        )

    def test_wrong_task_type(self, capsys: CaptureFixture):
        """See if a wrong file structure (not a dict) is handled correctly"""
        self._check_config_error(
            capsys, "test_wrong_task_type.yaml", "File does not define a dict"
        )

    def test_eventual_empty_config(self, capsys: CaptureFixture):
        """See if a custom config that results in an empty config (after tasks are disabled) is handled correctly"""
        self._check_config_error(
            capsys, "test_eventual_empty_config.yaml", "No valid tasks defined"
        )

    def test_nonexistent_task(self, capsys: CaptureFixture):
        """See if a task that does not exist in the default config is handled correctly"""
        self._check_config_warning(
            capsys,
            "test_nonexistent_task.yaml",
            "disabled because it is not defined in the default config",
        )

    def test_disabled_dependency(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys,
            "test_disabled_dependency.yaml",
            "disabled because it depends on disabled tasks",
        )

    ### Properties ####

    def test_extra_task_property(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys, "test_extra_task_property.yaml", self.fields_message
        )

    def test_wrong_task_property_type(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys, "test_wrong_task_property_type.yaml", self.fields_message
        )

    def test_wrong_args_type(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys,
            "test_wrong_args_type.yaml",
            "because 'args' field is not a dict",
        )

    ### Args ####

    def test_extra_task_arg(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys, "test_extra_task_arg.yaml", self.fields_message
        )

    def test_wrong_task_arg_type(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys, "test_wrong_task_arg_type.yaml", self.fields_message
        )

    def test_unallowed_arg(self, capsys: CaptureFixture):
        self._check_config_warning(
            capsys,
            "test_unallowed_arg.yaml",
            "disabled because it uses unallowed args",
        )



import os
import subprocess
import sys
from pathlib import Path

GROUP_CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "max_content_width": 200,
}  # optional: controls text wrapping
SUBCOMMAND_CONTEXT_SETTINGS = {
    "ignore_unknown_options": True,
    "allow_extra_args": True,
    "help_option_names": [],
}


def get_prog(info_name: str) -> str:
    """Build a prog string for argparse subcommands."""
    return f"{Path(sys.argv[0]).name} {info_name}"


def get_sudo_access() -> None:
    """Request sudo access from the user (unless help is asked)."""
    if not {"-h", "--help"} & set(sys.argv) and not os.geteuid() == 0:
        # Call the same script with sudo
        result = subprocess.call(["sudo", "-E", *sys.argv])
        sys.exit(result)
